Listing: MDX11.C

// MDX11.C: List out MDX structures by file name or wildcard
//          Written for Turbo C++ version 1.0, should compile
//          with any version of Turbo C.

#include <string.h>
#include <io.h>
#include <stdio.h>
#include <fcntl.h>
#include <dir.h>

#define byte unsigned char

#define FILENAMELENGTH 13
#define MAXFILES 100          // arbitrary, no/files in file table
#define YEAR 0  // --|
#define MON  1  //   | for YMD-format dates
#define DAY  2  // --|
#define MXKYEXP 220           // max length/index key expr.
#define MDXTAGLEN 10          // longest MDX tag

typedef struct {
    char version;           // =02
    char date[3];           // hex YMD
    char dbfnam[16];        // room for UNIX names??
    short blksiz;           // #/512-byte blocks/blk
    short bytsiz;           // #/bytes/block (blksiz*MDXBLKSIZE)
    byte  maintain;         // 0==not productn mdx 1==is
    byte  slotcnt;          // # slots in table
    short slotsiz;          // # bytes in slot
    byte  mdxcnt;           // #index slots used
    char  damaged;          // 1==damaged
    byte  btmodcnt;         // mu cnt./changes to tag  ??
    byte  spare;
    long  newblk;           // block # of next avail. blk
    long  freelst;          // block # of 1st recycled blk
    char  mystery[4];       // who knows?
    char  creatdate[3];     // hex, YMD
    char  pad2[465];        // pad to 512 bytes
} MDXHDR;

      // MDX "directory" entry
typedef struct {
    long blockno;           // physical block # of header
    char tag[MDXTAGLEN+1];  // ASCII name of INDEX
    byte type;              // SQL/dBASE index expression
    byte lslot,rslot;       // binary tree pointers for ASCII dir. order
    byte pslot;             // parent tree ptr
    byte flags;             // MDXINDX/NDXINDX
    byte dtype;             // C)HAR, F)LOAT, N)UMBER, D)ATE
    byte spares[11];        // spare space
} MDXSLOT;

      // individual entry in index header list of MDX file
typedef struct {
    long  btroot;           // block number of root
    long  btavail;          // # blocks in this chain
    char  btflags;          // bitmap, 8=descending,16=??
    char  bttype;           // C,N,F,D
    char  spares[2];
    short btkeylen;         // <= 100: length of key expressions
    short btmaxkey;         //
    short btnumber;         // ==1 if numeric key
    short entsiz;           // btkeylen + some overhead?
    byte  disk_mod_cnt;
    char  btcrypt;          // encrypted==1
    char  damaged;          // damaged (under cnstrct.)==1
    char  btunique;         // unique==nonzero
    char  btkeyst[MXKYEXP+1];// key expression
    char  spares2[3];
    long  endchain;         // last block in chain
    long  btrecycl;
    char  idunno1;
    char  creatdate[3];     // date of create or last REINDEX
    char  idunno2[502];     // seems to contain info in places, but what?
    char  for_expr[262];

} HEADER_ENTRY;

// messages for people who snoop in EXE files
static char copywrong[]="MDX11 Copywrong R. Freeland: no rights reserved";
static char whazzup[]="What are you doing snooping in here, anyway?";

// global data
short fhandle,                // file handle, only one file open @ a time
      tagcount=0;             // count of tags processed/current file
MDXHDR mdxhdr;                // header area to read into
MDXSLOT slots[48];            // directory table to read into
HEADER_ENTRY header_rec;      // MDX header entry to read into
struct ffblk dta;             // for Turbo C's find first/next
char filetable[MAXFILES][FILENAMELENGTH];   // stores filenames

short getfiles(char *paramfname, char *ext) {
    char wildcard[128];
    short numfiles=0,b,done;
    char *mark,*fname;

    strcpy(wildcard,paramfname);            // dump it to local
    if ((!strpbrk(wildcard,"\*\?")) && findfirst(wildcard,&dta,16)==0) {
          // if not a wildcard and is a valid directory name
        if (dta.ff_attrib==16) {            // append "\"
            strcat(wildcard,"\\");
            strcat(paramfname,"\\");
        }
    }
    fname=(strrchr(wildcard,'\\'));         // look for end of dirname
    if (fname==NULL) fname=strrchr(wildcard,':'); // then for drive
    if (fname==NULL) fname=wildcard; else fname++;
    if (strlen(fname)==0) strcpy(fname,"*");
    mark=strchr(fname,'.');
    if (mark==NULL) {                       // if no "." in name,
        strcat(wildcard,".");               // add ".MDX"
        strcat(wildcard,ext);
    }
    done=(findfirst(wildcard,&dta,39));     // find any file but directory
    if (done) {
        printf("%s: No matching files found!",wildcard);
        exit(1);
    }
    while (!done) {
        strcpy(filetable[numfiles++],dta.ff_name);  // save filenames
        if (numfiles==MAXFILES) {
            printf("Whoa! %d maximum files exceeded!\n",MAXFILES);
            break;
        }
        done=findnext(&dta);
    }
    if (numfiles>1) {
        qsort(filetable,numfiles,FILENAMELENGTH,strcmp);
        printf("%d files matching %s:\n",numfiles,wildcard);
    }
    return(numfiles);
}

// convert dBASE-header-format date (YMD) to character string
char *dtoc(char *retstring,char *dbdate) {
    sprintf(retstring,"%02d/%02d/%02d",dbdate[MON],dbdate[DAY],dbdate[YEAR]);
    return(retstring);
}

// go climb a tree: process array of slots as binary tree
void climb(byte curslot) {
    char datestring[9];

    if (curslot==0) return;
    climb(slots[curslot].lslot);        // go left before middle
    tagcount++;
    if (tagcount>mdxhdr.mdxcnt) {       // just in case [g]
        printf("Whoops . . . got lost in Btree!\n");
        return;
    }
    lseek(fhandle,512*slots[curslot].blockno,0);    // position to header
    _read(fhandle,&header_rec,sizeof(HEADER_ENTRY));// and read it in
    printf("%-10s  %c   %c %3d  %-12s  %c  %s %s\n",
        slots[curslot].tag,
        slots[curslot].dtype,
        header_rec.btflags&8?'D':'A',
        header_rec.btkeylen,
        mdxhdr.dbfnam,
        mdxhdr.maintain?'Y':'N',
        dtoc(datestring,header_rec.creatdate),
        header_rec.btkeyst
        );
    if (header_rec.damaged) printf(" *DAMAGED!\n");
    if (header_rec.btcrypt) printf(" **ENCRYPTED!\n");
    if (header_rec.btunique) printf(" ***UNIQUE!\n");
    if (header_rec.for_expr[0]!=0) printf(" ****CONDITION: %s\n",
                                              header_rec.for_expr);
    climb(slots[curslot].rslot);        // go right after middle
}

// cut out just the path from a path\filename, return path length
short justpath(char *anyfname) {
    char *mark;

    if ((mark=strrchr(anyfname,'\\'))!=NULL) *(mark+1)=0;
     else if ((mark=strchr(anyfname,':'))!=NULL) *(mark+1)=0;
      else *anyfname=0;
    return(strlen(anyfname));
}

void main(short argc, char *argv[]) {
    char pathname[128],datestring1[9],datestring2[9];
    short filecount, numfiles;
    short pathlen;

    _fmode=O_BINARY;
    if (argc < 2) strcpy(pathname,"*.MDX");
     else strcpy(pathname,argv[1]);
    numfiles=getfiles(pathname,"MDX");
    pathlen=justpath(pathname);
    for (filecount=0;filecount<numfiles;filecount++) {
        strcpy(pathname+pathlen,filetable[filecount]);
        printf("\n%s: ",pathname);
        if ((fhandle = open(pathname,O_RDONLY+O_BINARY)) == -1) {
            printf("Can't open file!\n");
            continue;
        }
        if (_read(fhandle,&mdxhdr,sizeof(MDXHDR))!=sizeof(MDXHDR)) {
            printf("Can't read header!\n");
            continue;
        }
        if (mdxhdr.version!=2) {
            printf("Unrecognized MDX version!\n");
            continue;
        }
        printf("DLUpdate=%s Creation date=%s\n",
            dtoc(datestring1,mdxhdr.date),
            dtoc(datestring2,mdxhdr.creatdate));
        _read(fhandle,slots,48*sizeof(MDXSLOT));
        tagcount=0;
        printf(
        "   Tag    Type A/D Len   DBF      Pdctn?  DLUpd  Expression\n");
        climb(slots[0].rslot);        // climb the tree of MDXSLOTs
        _close(fhandle);
    }
}
